"""
Logging utils
"""


import sys
import colorama


class Log(object):
    """
    Logging class
    """

    def __init__(self):
        colorama.init()
        self.quiet = False

    def set_quiet(self, quiet=True):
        self.quiet = quiet

    def print_color(self, color_code, msg):
        if self.quiet:
            return
        sys.stderr.write(color_code)
        sys.stderr.write(msg)
        sys.stderr.write(colorama.Style.RESET_ALL)
        sys.stderr.write("\n")
        sys.stderr.flush()

    def debug(self, msg):
        self.print_color(colorama.Fore.MAGENTA, msg)

    def info(self, msg):
        self.print_color(colorama.Fore.GREEN, msg)

    def warning(self, msg):
        self.print_color(colorama.Fore.YELLOW, msg)

    def error(self, msg):
        self.print_color(colorama.Fore.RED, msg)



"""
O2Tuner module
"""


import os
import yaml
from yaml import YAMLError
from o2tuner.log import Log


LOG = Log()


class O2TunerError(Exception):
    """
    O2Tuner error class
    """

    def __init__(self, msg):
        super().__init__()
        self.msg = msg

    def __str__(self):
        return self.msg


class O2Tuner(object):
    """
    Steering class for tuning
    """

    def __init__(self, handler) -> None:
        self._handler = handler
        self._conf = {}

    @staticmethod
    def get_default_conf():
        return {"executable": "/bin/true"}

    def parse_config(self):
        conf_file = os.path.join(os.path.expanduser("~"),
                                 ".o2tuner-config.yaml")
        try:
            with open(conf_file, encoding="utf8") as conf_file_data:
                conf_override = yaml.safe_load(conf_file_data)
                for k in self._conf:
                    self._conf[k] = conf_override.get(k, self._conf[k])
        except (OSError, IOError, YAMLError, AttributeError):
            pass

    def override_config(self, override):
        if not override:
            return
        for k in self._conf:
            if not override.get(k) is None:
                self._conf[k] = override[k]

    def reset_handler(self, handler):
        self._handler = handler

    def init(self, **kwargs):
        self._handler.initialise(**kwargs)

    def run(self):
        self._handler.optimise()



"""
Mainstream and custom backends should be placed here
"""


import optuna


class OptunaHandler(object):
    """
    Handler based on Optuna backend
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        self._objective = None
        self._study = None
        self._n_trials = 100

    def initialise(self, n_trials=100):
        self._n_trials = n_trials
        self._study = optuna.create_study()

    def optimise(self):
        self._study.optimize(self._objective, n_trials=self._n_trials)

    def set_objective(self, objective):
        self._objective = objective



"""
o2tuner module
"""

import sys
from pkg_resources import require
from o2tuner.argumentparser import O2TunerArgumentParser
from o2tuner.o2tuner import O2Tuner, O2TunerError
from o2tuner.backends import OptunaHandler
from o2tuner.log import Log


def objective(trial):
    x_var = trial.suggest_float("x", -10, 10)
    return (x_var-2)**2


LOG = Log()


def entrypoint():
    arg_parser = O2TunerArgumentParser()
    arg_parser.gen_config_help(O2Tuner.get_default_conf())
    args = arg_parser.parse_args()

    LOG.set_quiet(args.quiet)

    try:
        process_actions(args)
    except O2TunerError as exc:
        LOG.error(f"Cannot continue: {exc}")
        sys.exit(10)


def process_actions(args):
    if args.version:
        ver = str(require(__package__)[0].version)
        if ver == "LAST-TAG":
            ver = "development version"
        print(f"{__package__} {ver}")
        return

    optuna_handler = OptunaHandler()
    optuna_handler.set_objective(objective)

    # Create and run the tuner
    tuner = O2Tuner(optuna_handler)

    if args.action in ["run"]:
        process_run(tuner, args)
    else:
        assert False, "invalid action"


def process_run(o2_tuner, args):
    if args.action == "run":
        LOG.info("Running ...")
        o2_tuner.init(n_trials=50)
        o2_tuner.run()



"""
Module to process command line arguments
"""


import os.path
import argparse
from collections import namedtuple
import yaml

O2TunerArg = namedtuple("O2TunerArg", "option config descr")


class O2TunerArgumentParser(argparse.ArgumentParser):
    """
    Lightweight subclass of ArgumentParser implementing some features required to display the
    o2tuner help: addition of normal arguments and matching between command-line
    options and configuration file options
    """

    def __init__(self):
        self.args_normal = []
        super().__init__(formatter_class=argparse.RawTextHelpFormatter)
        super().add_argument("-v", "--version", dest="version", default=False, action="store_true",
                             help="Print current alidock version on stdout")
        super().add_argument("-q", "--quiet", dest="quiet", default=False,
                             action="store_true", help="Do not print any message")
        super().add_argument("-d", "--debug", dest="debug", default=None,
                             action="store_true", help="Increase verbosity")
        super().add_argument("action", default="run",
                             nargs="?", choices=["run"], help="Run optimiser")

    def gen_config_help(self, default_conf):
        conf_file = os.path.join(os.getcwd(), ".o2tuner-config.yaml")
        epilog = f"It is possible to specify the most frequently used options in a YAML " \
                 f"configuration file in your working directory\n" \
                 f"Current expected path is: {conf_file}\n" \
                 f"The following options (along with their default values) can be specified " \
                 f"(please include `---` as first line):\n---\n"
        yaml_lines = {}
        longest = 0
        for opt in self.args_normal:
            if opt.config:
                assert opt.config in default_conf, f"option {opt.config} expected in default conf"
                optd = {opt.config: default_conf[opt.config]}
                yaml_lines[opt.option] = yaml.dump(
                    optd, default_flow_style=False).rstrip()
                longest = max(longest, len(yaml_lines[opt.option]))
        fmt = f"%%-{longest}s  # same as option %%s\n"
        for y_line in yaml_lines.items():
            epilog += fmt % (yaml_lines[y_line], y_line)

        self.epilog = epilog
